package java.util.stream;

import java.util.Spliterator;
import java.util.EnumMap;
import java.util.Map;

enum StreamOpFlag
{
    DISTINCT(0, set(StreamOpFlag.Type.SPLITERATOR).set(StreamOpFlag.Type.STREAM).setAndClear(StreamOpFlag.Type.OP)), 
    SORTED(1, set(StreamOpFlag.Type.SPLITERATOR).set(StreamOpFlag.Type.STREAM).setAndClear(StreamOpFlag.Type.OP)), 
    ORDERED(2, set(StreamOpFlag.Type.SPLITERATOR).set(StreamOpFlag.Type.STREAM).setAndClear(StreamOpFlag.Type.OP).clear(StreamOpFlag.Type.TERMINAL_OP).clear(StreamOpFlag.Type.UPSTREAM_TERMINAL_OP)), 
    SIZED(3, set(StreamOpFlag.Type.SPLITERATOR).set(StreamOpFlag.Type.STREAM).clear(StreamOpFlag.Type.OP)), 
    SHORT_CIRCUIT(12, set(StreamOpFlag.Type.OP).set(StreamOpFlag.Type.TERMINAL_OP));
    
    private static final int SET_BITS = 1;
    private static final int CLEAR_BITS = 2;
    private static final int PRESERVE_BITS = 3;
    private final Map<StreamOpFlag.Type, Integer> maskTable;
    private final int bitPosition;
    private final int set;
    private final int clear;
    private final int preserve;
    static final int SPLITERATOR_CHARACTERISTICS_MASK;
    static final int STREAM_MASK;
    static final int OP_MASK;
    static final int TERMINAL_OP_MASK;
    static final int UPSTREAM_TERMINAL_OP_MASK;
    private static final int FLAG_MASK;
    private static final int FLAG_MASK_IS;
    private static final int FLAG_MASK_NOT;
    static final int INITIAL_OPS_VALUE;
    static final int IS_DISTINCT;
    static final int NOT_DISTINCT;
    static final int IS_SORTED;
    static final int NOT_SORTED;
    static final int IS_ORDERED;
    static final int NOT_ORDERED;
    static final int IS_SIZED;
    static final int NOT_SIZED;
    static final int IS_SHORT_CIRCUIT;
    
    private static StreamOpFlag.MaskBuilder set(StreamOpFlag.Type type) {
        return new StreamOpFlag.MaskBuilder(new EnumMap(StreamOpFlag.Type.class)).set(type);
    }
    
    private StreamOpFlag(int bitPosition, StreamOpFlag.MaskBuilder maskBuilder) {
        this.maskTable = maskBuilder.build();
        bitPosition *= 2;
        this.bitPosition = bitPosition;
        this.set = 1 << bitPosition;
        this.clear = 2 << bitPosition;
        this.preserve = 3 << bitPosition;
    }
    
    int set() {
        return this.set;
    }
    
    int clear() {
        return this.clear;
    }
    
    boolean isStreamFlag() {
        return (int)this.maskTable.get(StreamOpFlag.Type.STREAM) > 0;
    }
    
    boolean isKnown(int n) {
        return (n & this.preserve) == this.set;
    }
    
    boolean isCleared(int n) {
        return (n & this.preserve) == this.clear;
    }
    
    boolean isPreserved(int n) {
        return (n & this.preserve) == this.preserve;
    }
    
    boolean canSet(StreamOpFlag.Type type) {
        return ((int)this.maskTable.get(type) & 0x1) > 0;
    }
    
    private static int createMask(StreamOpFlag.Type type) {
        int n = 0;
        for (StreamOpFlag streamOpFlag : values()) {
            n |= (int)streamOpFlag.maskTable.get(type) << streamOpFlag.bitPosition;
        }
        return n;
    }
    
    private static int createFlagMask() {
        int n = 0;
        StreamOpFlag[] values = values();
        for (int length = values.length, i = 0; i < length; ++i) {
            n |= values[i].preserve;
        }
        return n;
    }
    
    private static int getMask(int n) {
        return (n == 0) ? StreamOpFlag.FLAG_MASK : (~(n | (StreamOpFlag.FLAG_MASK_IS & n) << 1 | (StreamOpFlag.FLAG_MASK_NOT & n) >> 1));
    }
    
    static int combineOpFlags(int n, int n2) {
        return (n2 & getMask(n)) | n;
    }
    
    static int toStreamFlags(int n) {
        return ~n >> 1 & StreamOpFlag.FLAG_MASK_IS & n;
    }
    
    static int toCharacteristics(int n) {
        return n & StreamOpFlag.SPLITERATOR_CHARACTERISTICS_MASK;
    }
    
    static int fromCharacteristics(Spliterator<?> spliterator) {
        int characteristics = spliterator.characteristics();
        if ((characteristics & 0x4) != 0x0 && spliterator.getComparator() != null) {
            return characteristics & StreamOpFlag.SPLITERATOR_CHARACTERISTICS_MASK & 0xFFFFFFFB;
        }
        return characteristics & StreamOpFlag.SPLITERATOR_CHARACTERISTICS_MASK;
    }
    
    static int fromCharacteristics(int n) {
        return n & StreamOpFlag.SPLITERATOR_CHARACTERISTICS_MASK;
    }
    
    static {
        SPLITERATOR_CHARACTERISTICS_MASK = createMask(StreamOpFlag.Type.SPLITERATOR);
        STREAM_MASK = createMask(StreamOpFlag.Type.STREAM);
        OP_MASK = createMask(StreamOpFlag.Type.OP);
        TERMINAL_OP_MASK = createMask(StreamOpFlag.Type.TERMINAL_OP);
        UPSTREAM_TERMINAL_OP_MASK = createMask(StreamOpFlag.Type.UPSTREAM_TERMINAL_OP);
        FLAG_MASK = createFlagMask();
        FLAG_MASK_IS = StreamOpFlag.STREAM_MASK;
        FLAG_MASK_NOT = StreamOpFlag.STREAM_MASK << 1;
        INITIAL_OPS_VALUE = (StreamOpFlag.FLAG_MASK_IS | StreamOpFlag.FLAG_MASK_NOT);
        IS_DISTINCT = StreamOpFlag.DISTINCT.set;
        NOT_DISTINCT = StreamOpFlag.DISTINCT.clear;
        IS_SORTED = StreamOpFlag.SORTED.set;
        NOT_SORTED = StreamOpFlag.SORTED.clear;
        IS_ORDERED = StreamOpFlag.ORDERED.set;
        NOT_ORDERED = StreamOpFlag.ORDERED.clear;
        IS_SIZED = StreamOpFlag.SIZED.set;
        NOT_SIZED = StreamOpFlag.SIZED.clear;
        IS_SHORT_CIRCUIT = StreamOpFlag.SHORT_CIRCUIT.set;
    }
    
    enum Type
    {
        SPLITERATOR, 
        STREAM, 
        OP, 
        TERMINAL_OP, 
        UPSTREAM_TERMINAL_OP;
    }
    
    private static class MaskBuilder
    {
        final Map<StreamOpFlag.Type, Integer> map;
        
        MaskBuilder(Map<StreamOpFlag.Type, Integer> map) {
            this.map = map;
        }
        
        StreamOpFlag.MaskBuilder mask(StreamOpFlag.Type type, Integer n) {
            this.map.put(type, n);
            return this;
        }
        
        StreamOpFlag.MaskBuilder set(StreamOpFlag.Type type) {
            return this.mask(type, Integer.valueOf(1));
        }
        
        StreamOpFlag.MaskBuilder clear(StreamOpFlag.Type type) {
            return this.mask(type, Integer.valueOf(2));
        }
        
        StreamOpFlag.MaskBuilder setAndClear(StreamOpFlag.Type type) {
            return this.mask(type, Integer.valueOf(3));
        }
        
        Map<StreamOpFlag.Type, Integer> build() {
            StreamOpFlag.Type[] values = StreamOpFlag.Type.values();
            for (int length = values.length, i = 0; i < length; ++i) {
                this.map.putIfAbsent(values[i], Integer.valueOf(0));
            }
            return this.map;
        }
    }
}
